////////////////////////////////////////////////////////////////////////////////
//
//  ****************************************************************************
//  * Unit Name : uDebuger
//  * Purpose   :   
//  * Author    :  (Rouse_) 
//  * Copyright :  Fangorn Wizards Lab 1998 - 2012.
//  * Version   : 1.0
//  * Home Page : http://rouse.drkb.ru
//  * Home Blog : http://alexander-bagel.blogspot.com/
//  ****************************************************************************
//

unit uDebuger;

interface

uses
  Windows, Classes, Controls, Forms, Dialogs, StdCtrls,
  ComCtrls, ExtCtrls, FWDebugerCore;

type
  TdlgDebuger = class(TForm)
    Panel1: TPanel;
    btnStart: TButton;
    btnStop: TButton;
    edLog: TRichEdit;
    procedure btnStartClick(Sender: TObject);
    procedure btnStopClick(Sender: TObject);
  private
    FCore: TFWDebugerCore;
    FNeedStop: Boolean;
    FStepCount: Integer;
    procedure OnCreateProcess(Sender: TObject; ThreadIndex: Integer;
      Data: TCreateProcessDebugInfo);
    procedure OnLoadDll(Sender: TObject; ThreadIndex: Integer;
      Data: TLoadDLLDebugInfo);
    procedure OnDebugString(Sender: TObject; ThreadIndex: Integer;
      Data: TOutputDebugStringInfo);
    procedure OnUnknownBreakPoint(Sender: TObject; ThreadIndex: Integer;
      ExceptionRecord: Windows.TExceptionRecord);
    procedure OnUnknownException(Sender: TObject; ThreadIndex: Integer;
      ExceptionRecord: Windows.TExceptionRecord);
    procedure OnIdle(Sender: TObject);
    procedure Writeln(const Value: string = '');
  end;

var
  dlgDebuger: TdlgDebuger;

implementation

uses SysUtils;

{$R *.dfm}

{ TdlgDebuger }

procedure TdlgDebuger.btnStartClick(Sender: TObject);
var
  Path: string;
begin
  FNeedStop := False;
  //       
  Path := ExtractFilePath(ParamStr(0)) + '..\test_app\test_app.exe';
  FCore := TFWDebugerCore.Create(50);
  try
    btnStart.Enabled := False;
    btnStop.Enabled := True;
    FCore.OnIdle := OnIdle;
    FCore.OnCreateProcess := OnCreateProcess;
    FCore.OnLoadDll := OnLoadDll;
    FCore.OnDebugString := OnDebugString;
    FCore.OnUnknownBreakPoint := OnUnknownBreakPoint;
    FCore.OnUnknownException := OnUnknownException;
    if not FCore.DebugNewProcess(Path, True) then
      RaiseLastOSError;
    FCore.RunMainLoop;
  finally
    FCore.Free;
    btnStart.Enabled := True;
    btnStop.Enabled := False;
  end;
  Writeln;
  Writeln('Debug stop');
end;

procedure TdlgDebuger.OnCreateProcess(Sender: TObject;
  ThreadIndex: Integer; Data: TCreateProcessDebugInfo);
var
  T: TThreadData;
begin
  T := FCore.GetThreadData(ThreadIndex);
  Writeln(Format('CreateThread ID: %d', [T.ThreadID]));
  Writeln(Format('ProcessStart ID: %d', [FCore.DebugProcessData.ProcessID]));
end;

procedure TdlgDebuger.OnDebugString(Sender: TObject; ThreadIndex: Integer;
  Data: TOutputDebugStringInfo);
begin
  if Data.fUnicode = 1 then
    Writeln('DebugString: ' + PWideChar(FCore.ReadStringW(Data.lpDebugStringData, Data.nDebugStringLength)))
  else
    Writeln('DebugString: ' + PAnsiChar(FCore.ReadStringA(Data.lpDebugStringData, Data.nDebugStringLength)));
end;

procedure TdlgDebuger.OnIdle(Sender: TObject);
begin
  if FNeedStop then
    FCore.StopDebug
  else
    Application.ProcessMessages;
end;

procedure TdlgDebuger.OnLoadDll(Sender: TObject; ThreadIndex: Integer;
  Data: TLoadDLLDebugInfo);
const
  FormatStrKnownDLL = 'Load Dll at instance %p handle %d "%s"';
  FormatStrUnknownDLL = 'Load unknown Dll at instance %p handle %d';
var
  DllName: AnsiString;
  IsUnicodeData: Boolean;
begin
  FCore.ContinueStatus := DBG_EXCEPTION_NOT_HANDLED;
  IsUnicodeData := Data.fUnicode = 1;
  DllName := FCore.GetDllName(Data.lpImageName, Data.lpBaseOfDll, IsUnicodeData);
  if DllName <> '' then
  begin
    if IsUnicodeData then
      Writeln(Format(FormatStrKnownDLL, [Data.lpBaseOfDll,
        Data.hFile, PWideChar(@DllName[1])]))
    else
      Writeln(Format(FormatStrKnownDLL, [Data.lpBaseOfDll,
        Data.hFile, PAnsiChar(@DllName[1])]));
  end
  else
    Writeln(Format(FormatStrUnknownDLL, [Data.lpBaseOfDll, Data.hFile]));
end;

procedure TdlgDebuger.OnUnknownBreakPoint(Sender: TObject;
  ThreadIndex: Integer; ExceptionRecord: Windows.TExceptionRecord);
var
  ApplicationBP: Boolean;
begin
  ApplicationBP :=
    (DWORD(ExceptionRecord.ExceptionAddress) > FCore.DebugProcessData.EntryPoint) and
    (DWORD(ExceptionRecord.ExceptionAddress) < $500000);

  Writeln;
  if ApplicationBP then
  begin
    Writeln(Format('!!! --> Unknown application breakpoint at addr 0X%p',
      [ExceptionRecord.ExceptionAddress]));
    Writeln('!!! --> Exception not handled.');
    FCore.ContinueStatus := DBG_EXCEPTION_NOT_HANDLED;
  end
  else
  begin
    Writeln(Format('!!! --> Unknown breakpoint at addr 0X%p',
      [ExceptionRecord.ExceptionAddress]));
    Writeln('!!! --> Exception handled.');
    FCore.ContinueStatus := DBG_CONTINUE;
  end;
  Writeln;
end;

procedure TdlgDebuger.OnUnknownException(Sender: TObject;
  ThreadIndex: Integer; ExceptionRecord: Windows.TExceptionRecord);
var
  Cause: string;
begin
  Writeln;

  case ExceptionRecord.ExceptionCode of

    EXCEPTION_STACK_OVERFLOW:
    begin
      Writeln('!!! --> Stack overflow detected. Probe to continue.');
      FCore.ContinueStatus := DBG_CONTINUE;
    end;

    EXCEPTION_ACCESS_VIOLATION:
    begin
      {
        The first element of the array contains a read-write flag that
        indicates the type of operation that caused the access violation.
        If this value is zero, the thread attempted to read the inaccessible data.
        If this value is 1, the thread attempted to write to an inaccessible address.
        If this value is 8, the thread causes a user-mode data execution prevention (DEP) violation.
        The second array element specifies the virtual address of the inaccessible data.
      }
      case ExceptionRecord.ExceptionInformation[0] of
        0: Cause := 'read';
        1: Cause := 'write';
        8: Cause := 'DEP violation';
      else
        Cause := 'unknown cause';
      end;

      Writeln(Format('!!! --> Access violation at addr 0x%p %s of address 0x%p',
        [
          ExceptionRecord.ExceptionAddress,
          Cause,
          Pointer(PDWORD(@ExceptionRecord.ExceptionInformation[1])^)
        ]));
      Writeln('!!! --> Process Stopped.');
      FCore.ContinueStatus := DBG_CONTROL_C;
    end;

  else

    Writeln(Format('!!! --> Unknown exception code %p at addr 0x%p',
      [
        Pointer(ExceptionRecord.ExceptionCode),
        ExceptionRecord.ExceptionAddress
      ]));

  end;

  Writeln;
end;

procedure TdlgDebuger.Writeln(const Value: string);
begin
  edLog.Lines.Add(Value);
end;

procedure TdlgDebuger.btnStopClick(Sender: TObject);
begin
  FNeedStop := True;
end;

end.
